open Okweb
open Core
open Lwt.Infix

type params = {
  entry_repo : Entry_repo.t;
  entry_template_repo : Entry_template_repo.t;
}

let get_months_of year dates_with_entries =
  let map_month month =
    let days_max = Date.days_in_month ~year ~month in
    let shift =
      Date.create_exn ~y:year ~m:month ~d:1
      |> Date.day_of_week |> Day_of_week.iso_8601_weekday_number |> Int.pred
    in
    let weeks_num =
      Float.of_int (days_max + shift) /. 7.0 |> Float.iround_up_exn
    in
    let map_week_day week_n day_n =
      match (week_n * 7) + day_n - shift + 1 with
      | day_of_month when day_of_month > 0 && day_of_month <= days_max ->
          let full_date = Date.create_exn ~y:year ~m:month ~d:day_of_month in
          `O
            [
              ("text", `String (Int.to_string day_of_month));
              ( "date",
                `String
                  (Printf.sprintf "%04d%02d%02d" year (Month.to_int month)
                     day_of_month) );
              ("has_entry", `Bool (Hash_set.mem dates_with_entries full_date));
            ]
      | _ -> `Null
    in
    let map_week week_n =
      let days = List.range 0 7 |> List.map ~f:(map_week_day week_n) in
      `O [ ("days", `A days) ]
    in
    let weeks = List.range 0 weeks_num |> List.map ~f:map_week in
    `O [ ("name", `String (Month.to_string month)); ("weeks", `A weeks) ]
  in
  `A (List.map Month.all ~f:map_month)

let get_calendar { entry_repo; _ } user_id year =
  let dates_with_entries =
    Entry_repo.find_dates_with_entries entry_repo ~user_id ~year
  in
  let template = Mustache.of_string [%blob "web/calendar_page.html"] in
  let params =
    `O
      [
        ("prev_year", `String Int.(to_string (pred year)));
        ("this_year", `String (Int.to_string year));
        ("next_year", `String Int.(to_string (succ year)));
        ("months", get_months_of year dates_with_entries);
      ]
  in
  let body = Mustache.render template params in
  Response.(
    ok
    |> with_header ~key:"Content-Type" ~value:"text/html;charset=utf-8"
    |> with_string_body body)

let get_entry { entry_repo; _ } user_id date entry_template_opt =
  let template = Mustache.of_string [%blob "web/entry_page.html"] in
  let entry_template = Option.value_exn entry_template_opt in
  let entry_content =
    Entry_repo.get_by_user_id_and_date entry_repo ~user_id ~date
    |> Option.value_map ~default:entry_template ~f:(fun e -> e.Entry.content)
  in
  let params =
    `O
      [
        ( "prev_date",
          `String (Date.to_string_iso8601_basic (Date.add_days date (-1))) );
        ( "next_date",
          `String (Date.to_string_iso8601_basic (Date.add_days date 1)) );
        ("date", `String (Date.format date "%b %e, %Y"));
        ("content", `String entry_content);
      ]
  in
  let body = Mustache.render template params in
  Response.(
    ok
    |> with_header ~key:"Content-Type" ~value:"text/html;charset=utf-8"
    |> with_string_body body)

let save_entry { entry_repo; _ } user_id date content =
  ignore (Entry_repo.create_or_update entry_repo ~user_id ~date ~content);
  Response.redirect
    (Printf.sprintf "/entry/%s" (Date.to_string_iso8601_basic date))

let get_first_time_page () =
  Response.(
    ok
    |> with_header ~key:"Content-Type" ~value:"text/html;charset=utf-8"
    |> with_string_body [%blob "web/first_time_page.html"])

let save_entry_template { entry_template_repo; _ } user_id content =
  ignore (Entry_template_repo.create entry_template_repo ~user_id ~content);
  Response.redirect "/calendar"

let make_error_page ~status ~title ~message =
  let template = Mustache.of_string [%blob "web/error_page.html"] in
  let params = `O [ ("title", `String title); ("message", `String message) ] in
  let body = Mustache.render template params in
  Response.(
    of_status status
    |> with_header ~key:"Content-Type" ~value:"text/html;charset=utf-8"
    |> with_string_body body)

let get_entry_template { entry_template_repo; _ } user_id =
  Entry_template_repo.get_by_user_id entry_template_repo ~user_id
  |> Option.map ~f:(fun et -> et.Entry_template.content)

let do_process_request params req =
  let user_id_opt =
    Request.get_cookies req |> Cookie_jar.find_value ~key:"user_id"
  in
  let entry_template_opt =
    user_id_opt |> Option.bind ~f:(get_entry_template params)
  in
  match Request.(get_http_method req, get_path req, user_id_opt) with
  (* Static routes *)
  | "GET", [ "okdiary.css" ], _ ->
      Lwt.return
        Response.(
          ok
          |> with_header ~key:"Content-Type" ~value:"text/css"
          |> with_string_body [%blob "web/okdiary.css"])
  | "GET", [ "okdiary.js" ], _ ->
      Lwt.return
        Response.(
          ok
          |> with_header ~key:"Content-Type" ~value:"text/javascipt"
          |> with_string_body [%blob "web/okdiary.js"])
  (* First time dialog *)
  | "GET", [ "first_time" ], Some _ -> Lwt.return (get_first_time_page ())
  | "POST", [ "entry_template" ], Some user_id -> (
      Request.get_body req |> Form.of_body >|= Form.find_param ~key:"content"
      >|= fun content_opt ->
      match content_opt with
      | Some content -> save_entry_template params user_id content
      | _ ->
          make_error_page ~status:400 ~title:"Invalid request"
            ~message:
              "The request you've sent does not contain required parameters" )
  | "GET", _, Some _ when Option.is_none entry_template_opt ->
      Lwt.return (Response.redirect "/first_time")
  (* Calendar *)
  | "GET", [], _ -> Lwt.return (Response.redirect "/calendar")
  | "GET", [ "calendar" ], Some user_id ->
      let current_year = Date.today ~zone:Time.Zone.utc |> Date.year in
      Lwt.return (get_calendar params user_id current_year)
  | "GET", [ "calendar"; year_str ], Some user_id ->
      Lwt.return
        ( match int_of_string_opt year_str with
        | Some year when 1900 < year && year < 9999 ->
            get_calendar params user_id year
        | _ ->
            make_error_page ~status:400 ~title:"Invalid year"
              ~message:"A year should be between 1900 and 9999." )
  (* Entry *)
  | "GET", [ "entry"; date_str ], Some user_id ->
      let date_opt =
        Option.try_with (fun () -> Date.of_string_iso8601_basic date_str ~pos:0)
      in
      Lwt.return
        ( match date_opt with
        | Some date when Int.equal (String.length date_str) 8 ->
            get_entry params user_id date entry_template_opt
        | _ ->
            make_error_page ~status:400 ~title:"Invalid date"
              ~message:"A date should be valid and in the format 'YYYYMMDD'" )
  | "POST", [ "entry"; date_str ], Some user_id -> (
      let date_opt =
        Option.try_with (fun () -> Date.of_string_iso8601_basic date_str ~pos:0)
      in
      Request.get_body req |> Form.of_body >|= Form.find_param ~key:"content"
      >|= fun content_opt ->
      match (date_opt, content_opt) with
      | Some date, Some content -> save_entry params user_id date content
      | _ ->
          make_error_page ~status:400 ~title:"Invalid request"
            ~message:
              "The request you've sent does not contain required parameters" )
  (* Errors *)
  | _ ->
      Lwt.return
        (make_error_page ~status:404 ~title:"Not found"
           ~message:"The page you're looking for is not found")

let process_request params = Looksok.middleware (do_process_request params)
