open Core
open Okweb

module Schema : sig
  val migrations : string list
end

module Entry : sig
  type t = {
    user_id : String.t;
    date : Date.t;
    content : String.t;
    created_at : Time.t;
    updated_at : Time.t;
  }
  [@@deriving equal, sexp]

  include Oksql.From_sql.Sig with type t := t

  val to_string : t -> string
end

module Entry_template : sig
  type t = { user_id : String.t; content : String.t }

  include Oksql.From_sql.Sig with type t := t
end

module Entry_repo : sig
  type t

  (** Initializes new entry repo using given Sqlite3 DB instance *)
  val init : Sqlite3.db -> t

  (** Creates new entry and inserts it or updates existing one in DB *)
  val create_or_update :
    user_id:string -> date:Core.Date.t -> content:string -> t -> Entry.t

  (** Finds entry written by the given user at the given date *)
  val get_by_user_id_and_date :
    user_id:string -> date:Core.Date.t -> t -> Entry.t option

  (** For the given [user_id] and [year] returns a set of dates that have
      non-empty diary entries associated with them. *)
  val find_dates_with_entries :
    user_id:string -> year:int -> t -> Core.Date.t Base.Hash_set.t
end

module Entry_template_repo : sig
  type t

  (** Initializes new entry template repo using given Sqlite3 DB instance *)
  val init : Sqlite3.db -> t

  (** Creates new entry template *)
  val create : user_id:string -> content:string -> t -> Entry_template.t

  (** Finds entry template by the given [user_id] *)
  val get_by_user_id : user_id:string -> t -> Entry_template.t option
end

module Daemon : sig
  type params = {
    entry_repo : Entry_repo.t;
    entry_template_repo : Entry_template_repo.t;
  }

  val process_request : params -> Handler.t
end
