const PASSWORD_KEY = 'password';

let ORIGINAL_ENTRY_CONTENT = '';

const $ = (selector) =>
  document.querySelector(selector);
const $entry = () =>
  $('#entry');
const $entryForm = () =>
  $('#entry form');
const $entryEncryptedContent = () =>
  $('#entry input[name="content"]');
const $editor = () =>
  $('#editor');
const $passwordPrompt = () =>
  $('#password-prompt');
const $passwordPromptForm = () =>
  $('#password-prompt form');
const $passwordInput = () =>
  $('#password-prompt input[type="password"]');
const $calendarDay = (date) =>
  $(`#calendar div[data-date="${date}"]`);
const $firstTimeForm = () =>
  $('#first-time-form');
const $firstTimeFormPasswordInput = () =>
  $('#first-time-form input[type="password"]');
const $firstTimeFormEncryptedContent = () =>
  $('#first-time-form input[name="content"]');

const getPassword = () =>
  window.sessionStorage.getItem(PASSWORD_KEY);

const putPassword = (password) =>
  window.sessionStorage.setItem(PASSWORD_KEY, password);

const deriveKey = async (password, salt) => {
  const material = await window.crypto.subtle.importKey(
    "raw",
    new TextEncoder().encode(password),
    {name: "PBKDF2"},
    false,
    ["deriveBits", "deriveKey"]);
  return window.crypto.subtle.deriveKey(
    {name: "PBKDF2", salt: salt, iterations: 65536, hash: "SHA-256"},
    material,
    {name: "AES-GCM", length: 256},
    true,
    ["encrypt", "decrypt"]);
}

const encrypt = async (plaintext, password) => {
  const getRandomValues = (len) =>
    window.crypto.getRandomValues(new Uint8Array(len));
  const writeBlobToDataUrl = async (blob) =>
    new Promise((resolve, reject) => {
      const r = new FileReader();
      r.onload = () => {
        resolve(r.result);
      };
      r.onerror = reject;
      r.readAsDataURL(blob);
    });
  const salt = getRandomValues(16);
  const key = await deriveKey(password, salt);
  const iv = getRandomValues(12);
  const ciphertext = await window.crypto.subtle.encrypt(
    {name: "AES-GCM", iv: iv},
    key,
    new TextEncoder().encode(plaintext));
  const blob = new Blob(
    [salt, iv, ciphertext],
    {type: 'application/octet-stream'});
  return writeBlobToDataUrl(blob);
}

const decrypt = async (encryptedDataUrl, password) => {
  const blob = await fetch(encryptedDataUrl).then(r => r.blob());
  const sliceBlobToArray = async (start, end) =>
    blob.slice(start, end).arrayBuffer().then(b => new Uint8Array(b));
  const salt = await sliceBlobToArray(0, 16);
  const iv = await sliceBlobToArray(16, 28);
  const ciphertext = await sliceBlobToArray(28);
  const key = await deriveKey(password, salt);
  const plaintext = await window.crypto.subtle.decrypt(
    {name: "AES-GCM", iv: iv},
    key,
    ciphertext)
  return new TextDecoder().decode(plaintext);
}

const autoGrow = (editor) => {
  if (editor.clientHeight < editor.scrollHeight) {
    editor.style.height = editor.scrollHeight + 'px';
  }
}

const editorTextChanged = () => {
  return ORIGINAL_ENTRY_CONTENT !== $editor().value;
}

const decryptEntry = async () => {
  const entryDataUrl = $entry().dataset.content;
  if (entryDataUrl !== '') {
    try {
      const editor = $editor();
      const entryContent = await decrypt(entryDataUrl, getPassword())
      ORIGINAL_ENTRY_CONTENT = entryContent;
      editor.setRangeText(entryContent, 0, editor.value.length);
      autoGrow(editor);
    } catch (e) {
      console.warn(e);
      askForPassword();
    }
  }
}

const encryptEntry = async () =>
  encrypt($editor().value, getPassword());

const setupEditor = () => {
  const editor = $editor();
  editor.addEventListener('input', () => autoGrow(editor));

  $entryForm().addEventListener('submit', async (event) => {
    event.preventDefault();
    $entryEncryptedContent().value = await encryptEntry();
    $entryForm().submit();
  });
}

const setupPasswordPrompt = () =>
  $passwordPromptForm().addEventListener('submit', (event) => {
    event.preventDefault();
    $passwordPrompt().style.display = 'none';
    putPassword($passwordInput().value);
    decryptEntry();
  });

const preventFromExitingWithUnsavedChanges = () =>
  window.addEventListener('beforeunload', (event) => {
    if (editorTextChanged() && $entryEncryptedContent().value === '') {
      event.preventDefault();
      event.returnValue = '';
    }
  });

const askForPassword = () => {
  $passwordPrompt().style.display = 'grid';
  $passwordInput().focus();
}

const highlightCurrentDay = () => {
  const now = new Date();
  const dd = String(now.getDate()).padStart(2, '0');
  const mm = String(now.getMonth() + 1).padStart(2, '0');
  const yyyy = String(now.getFullYear());
  const today = yyyy + mm + dd;
  $calendarDay(today).classList.toggle('today');
}

const onLoadEntryPage = () => {
  setupPasswordPrompt();
  setupEditor();
  preventFromExitingWithUnsavedChanges();
  if (getPassword() === null) {
    askForPassword();
  } else {
    decryptEntry();
  }
};

const onLoadCalendarPage = () =>
  highlightCurrentDay();

const onLoadFirstTimePage = () => {
  const editor = $editor();
  editor.addEventListener('input', () => autoGrow(editor));

  $firstTimeForm().addEventListener('submit', async (event) => {
    event.preventDefault();
    const password = $firstTimeFormPasswordInput().value;
    if (password !== "") {
      putPassword(password);
      $firstTimeFormEncryptedContent().value = await encrypt($editor().value, password);
      $firstTimeForm().submit();
    }
  });
};
