open Core
open Syslog
open Okweb
open Okdiary

let from_env ~conv ~key ~default =
  try Sys.getenv key |> Option.map ~f:conv |> Option.value ~default
  with _ ->
    sprintf "Warning: invalid env variable: %s. Using default value." key
    |> syslog;
    default

let from_env_or_abort ~conv ~key =
  let value_opt = Sys.getenv key in
  try
    Option.value_exn
      ( value_opt
      |> Option.filter ~f:(fun s -> not (String.is_empty s))
      |> Option.map ~f:conv )
  with _ ->
    sprintf "Error: invalid or absent env variable: %s. Exiting." key |> syslog;
    exit 1

let get_okauth_config () =
  let open Okcrypto in
  let open Okauth_client in
  let base_url = from_env_or_abort ~key:"BASE_URL" ~conv:Fn.id in
  let app_name = "OkDiary" in
  let public_encryption_key =
    from_env_or_abort ~key:"PUBLIC_ENCRYPTION_KEY"
      ~conv:Public_key.of_base64_exn
  in
  let secret_encryption_key =
    from_env_or_abort ~key:"SECRET_ENCRYPTION_KEY"
      ~conv:Secret_key.of_base64_exn
  in
  let public_signing_key =
    from_env_or_abort ~key:"PUBLIC_SIGNING_KEY" ~conv:Public_key.of_base64_exn
  in
  let secret_signing_key =
    from_env_or_abort ~key:"SECRET_SIGNING_KEY" ~conv:Secret_key.of_base64_exn
  in
  let storage = In_memory_storage.empty in
  Config.make ~base_url ~app_name ~public_encryption_key ~secret_encryption_key
    ~public_signing_key ~secret_signing_key storage

let dummy_auth_middleware handler req =
  let jar_with_test_user =
    Request.get_cookies req |> Cookie_jar.set (Cookie.make ("user_id", "dummy"))
  in
  let req_with_test_user = Request.with_cookies jar_with_test_user req in
  handler req_with_test_user

let serve () =
  openlog ~id:"okdiaryd" ~options:[ Open_option.PID; Open_option.PERROR ] ();
  let on_exn ex =
    sprintf "Unhandled exception: %s" (Exn.to_string ex) |> syslog
  in
  let address, port, db_path, auth_backend =
    ( from_env ~key:"HOST" ~default:"localhost" ~conv:Fn.id,
      from_env ~key:"PORT" ~default:4201 ~conv:Int.of_string,
      from_env ~key:"DB_PATH" ~default:"okdiary.db" ~conv:Fn.id,
      from_env ~key:"AUTH_BACKEND" ~default:"okauth" ~conv:Fn.id )
  in
  let auth_middleware =
    match auth_backend with
    | "okauth" ->
        let module Auth =
          Okauth_client.Make (Okweb.Client) (Okauth_client.In_memory_storage)
        in
        Auth.middleware (get_okauth_config ())
    | "dummy" -> dummy_auth_middleware
    | _ ->
        sprintf
          "Error: invalid authentication backend: %s. Valid options are: \
           okauth, dummy. Exiting."
          auth_backend
        |> syslog;
        exit 1
  in
  Oksql.with_db db_path (fun db ->
      Oksql.Migration.apply db ~migrations:Schema.migrations;
      let entry_repo = Entry_repo.init db in
      let entry_template_repo = Entry_template_repo.init db in
      let handler =
        Daemon.(process_request { entry_repo; entry_template_repo })
      in
      sprintf "Starting OkDiary Daemon v%s on %s:%d" [%blob "../../version"]
        address port
      |> syslog;
      Lwt_main.run
        (Server.listen ~address ~port ~on_exn (auth_middleware handler)))

let generate_application_keys () =
  let open Okcrypto in
  let aepk, aesk = generate_keypair Key_purpose.Encryption in
  let aspk, assk = generate_keypair Key_purpose.Signing in
  printf "PUBLIC_ENCRYPTION_KEY=\"%s\"\n" (Public_key.to_base64 aepk);
  printf "SECRET_ENCRYPTION_KEY=\"%s\"\n" (Secret_key.to_base64 aesk);
  printf "PUBLIC_SIGNING_KEY=\"%s\"\n" (Public_key.to_base64 aspk);
  printf "SECRET_SIGNING_KEY=\"%s\"\n" (Secret_key.to_base64 assk)

let () =
  match Sys.get_argv () with
  | [| _; "serve" |] -> serve ()
  | [| _; "generate-application-keys" |] -> generate_application_keys ()
  | [| _ |] ->
      Caml.prerr_endline
        "No command given. See 'man 8 okdiaryd' for available commands"
  | argv ->
      let drop_fst l = List.drop l 1 in
      let args = argv |> Array.to_list |> drop_fst |> String.concat ~sep:" " in
      eprintf
        "Invalid arguments: %s. See 'man 8 okdiaryd' for available commands \
         and arguments\n"
        args
